@@ -25,6 +25,7 @@ require 'kconv';
 require 'iconv';
 require 'fileutils';
 require 'pstore';
+require 'date';
 
 require 'bigdecimal';
 require 'bigdecimal/util';
@@ -66,8 +67,27 @@ AmConf['DirtyMailDir'] = AmConf['MailHome'] + '/dirty';
 
 # =====================================================================
 #
-##AmConf['DontUpdateTokenDB'] = 'yes';
-		# ǤʤʸʤСǡ١򹹿ʤ
+##AmConf['DontUpdateTokenDB'] = 'on';
+		# ǡ١򹹿ʤ
+
+##AmConf['AcceptHtml'] = 'on';
+		# "text/html" ʬޤȽꤹ롣
+		#	դξ硢ΤHTMLΤߤΥ᡼ϡֶԤʤפ
+		#	Ȥˤʤꡢä
+		#		AmConf['SpamIfEmptyBody'] = 'on'
+		#	ʤХѥȽǤ롣
+
+# =====================================================================
+#
+AmConf['SpamIfEmptyBody'] = 'on';
+		# ʸ˶Ԥʤ硢ѥȴ
+AmConf['SpamIfIllegalDate'] = 'on';
+		#  (Date:) ʤ硢ѥȴ
+		# -- ʸȤƲԲ
+		# -- Ȥκ1.5ʾ
+
+AmConf['SpamCharsets'] = 'utf-7,unicode-1-1-utf-7,gb2312,gb18030,koi8,koi8-r,koi8-u';
+		# 沽ѤƤΤϥѥȴ
 
 # =====================================================================
 #
@@ -75,7 +95,7 @@ AmConf['MailBoxFile'] = '/var/mail/%user%';
 		# ѥȽꤷʤäȤΥեɵ롣
 		# â줬ǤС$stdout˽񤭽Ф
 
-AmConf['LockFile'] = '/tmp/FEML_LOCK_FILE.txt';
+AmConf['LockFile'] = '/tmp/FEML_LOCK_FILE.%user%.txt';
 
 # =====================================================================
 #
@@ -84,7 +104,15 @@ AmConf['SpamThreshold'] = 0.95;
 # ---------------------------------------------------------------------
 # ---------------------------------------------------------------------
 #
-FemlVersion = 'feml-0.2.0';
+FemlVersion = 'feml-0.3.0';
+
+# =====================================================================
+# ֡뷿ܤˤĤơˤʤäƤ뤫ɤĴ٤롣
+#	ͤȤƶǤʤʸꤵƤХ
+#
+def is_conf_on(key)
+	return (AmConf[key] != nil && AmConf[key] != '');
+end
 
 # =====================================================================
 # ѥΨȽ
@@ -125,6 +153,7 @@ class SpamProb
 			%r|</(#{tags})>|im,
 			%r|<(#{tags})/>|im,
 			%r|<\?>|im,
+			%r|<feml.+?>|im,
 		];
 		patterns.each do |pattern|
 			src = src.gsub(pattern, ' ');
@@ -488,7 +517,9 @@ class ExDecoder
 			0xDB => 'U',		# &Ucirc;
 			0xDC => 'U',		# &Uuml;
 			0xDD => 'Y',		# &Yacute;
-			0xDE => '$',		# &THORN;
+			0xDE => 'P',		# &THORN;
+						# PפѤȤƻȤäƤ
+						# 㤬¿褦˸롣
 			0xDF => 'ss',		# &szlig;
 			0xE0 => 'a',		# &agrave;
 			0xE1 => 'a',		# &aacute;
@@ -520,7 +551,7 @@ class ExDecoder
 			0xFB => 'u',		# &ucirc;
 			0xFC => 'u',		# &uuml;
 			0xFD => 'y',		# &yacute;
-			0xFE => '$',		# &thorn;
+			0xFE => 'p',		# &thorn;
 			0xFF => 'y',		# &yuml;
 		};
 
@@ -586,15 +617,26 @@ class ExDecoder
 			from_enc = 'ISO-2022-JP-2';
 			mbuf = decode_by_iconv(mbuf, 'EUC-JP', from_enc);
 
-		when 'gb2312'				# λ
-			from_enc = Encoding::GBK;		# gb2312γĥ
+		when 'gb2312', 'gb18030',		# λ
+		     'windows-936', 'euc-cn',
+		     'x-euc'
+			from_enc = Encoding::GB18030;
 			mbuf = mbuf.encode(Encoding::EUC_JP, from_enc,
 				:invalid => :replace,
 				:undef => :replace,
 				:replace => '<?>');
-
-		# 'windows-936'				# λGBK
-		# 'gb18030'				# GBKξ̽??
+					#
+					# x-euc (cn/jp/krʤɤĤƤʤ) 
+					# ɤ٤뤬Ȥ㤬
+					# GBKȤƤޤѴǤΤǡ
+					# νǤФ餯ͻҤ򸫤롣
+					#
+			## mbuf = decode_by_iconv(mbuf, 'EUC-JP', 'GB18030');
+					#
+					# iconvѴ̤ϤȤɤ
+					# ǤϤʤȻפս꤬Τǡ
+					# String.encode()Ȥ
+					#
 
 		# 'windows-949'				# Korean
 		# 'windows-950'				# λBig5
@@ -621,11 +663,17 @@ class ExDecoder
 		     'iso-8859-13',			# Latin-7 (Хȸ)
 		     'iso-8859-14', 'windows-1257',	# Latin-8 (ȸ)
 		     'iso-8859-15',			# Latin-9 (8859-1)
-		     'iso-8859-16'			# Latin-10 (첤)
+		     'iso-8859-16',			# Latin-10 (첤)
+		     'us-ascii'				# us-ascii 7-bit
 					# ºݤˤϤ줾ʸγƤ
 					# ۤʤ뤬
 					# ѴΤŪǤ
 					# ʤΤǡwindows-1252Ѥ롣
+					#
+					# us-ascii (7-bit) ȤʤäƤƤ
+					# ºݤˤ[\x80-\xFF]ΥХȤ
+					# äƤ㤬ΤǤѴ
+					#
 			mbuf = windows_1252_to_ascii(mbuf);
 
 		# 'iso-8859-6', 'windows-1256'		# ӥ
@@ -636,7 +684,6 @@ class ExDecoder
 		# 'iso-8859-12'			# ǡʡ꡼(˴)
 
 		# else
-		# 'us-ascii'				#
 		# 'windows-932'				# Japanese Shift-JIS
 		# 'shift_jis'				#
 		# 'shift-jis'				#
@@ -765,15 +812,25 @@ class ExDecoder
 			end
 		end
 
+		# -----------------------------------------------------
+		# multipartζԤĴ٤롣
+		#
 		@boundary_lines = Array.new();
 		if @boundary != nil && @boundary != ''
 			b_regex = Regexp.new("--#{Regexp.escape(@boundary)}(--)?$");
+			end_boundary_appeared = false;
 			@mbuf_body_array.each_index do |i|
 				mbuf = @mbuf_body_array[i].toeuc;
 				if b_regex =~ mbuf
+					end_boundary_appeared = true if $1 == '--';
 					@boundary_lines << i;
 				end
 			end
+			if ! end_boundary_appeared
+				@boundary_lines << @mbuf_body_array.size;
+					# 򼨤ԤƤ
+					# 䤦
+			end
 		end
 
 ##		print "[encoding] #{@encoding}\n";
@@ -786,6 +843,15 @@ class ExDecoder
 	end
 	private :parse_header;
 
+	# =============================================================
+	# إåƤ; Ȥmultipartξꡣ
+	#
+	def header_summary_line()
+		return <<-"__SUMMARY__";
+## <feml content-type="#{@content_type_major}/#{@content_type_minor}" charset="#{@charset}" encoding="#{@encoding}" />
+		__SUMMARY__
+	end
+
 	# -------------------------------------------------------------
 	# إåλϤˤĤ:
 	#
@@ -832,6 +898,26 @@ class ExDecoder
 	end
 	protected :each_multipart;
 
+	# -------------------------------------------------------------
+	# Content-TypeȤơȽǺȤƻȤݤĴ٤롣
+	#
+	def content_type_is_acceptable()
+		is_acceptable = false;
+		case @content_type_major
+		when 'text'
+			case @content_type_minor
+			when 'plain'
+				is_acceptable = true;
+			when 'html'
+				is_acceptable = true if is_conf_on('AcceptHtml');
+			end
+		when 'message'
+			is_acceptable = true;
+		end
+		return is_acceptable;
+	end
+	private :content_type_is_acceptable;
+
 	# -------------------------------------------------------------
 	# ʸϤ롣
 	# - multipartξ硢Content-TypetextǤʤ󥯤Ϻ
@@ -862,40 +948,38 @@ class ExDecoder
 
 		elsif @boundary_lines.size != 0
 			self.each_multipart do |mbuf_array|
-##				@mbuf_body += "--#{@boundary}\n";
-				@mbuf_body += "----------\n";
 				sub_ex_d = ExDecoder.new();
 				sub_ex_d.load_array(mbuf_array);
+				@mbuf_body += %Q|## <feml boundary="#{@boundary}" />\n|;
+				@mbuf_body += sub_ex_d.header_summary_line;
 				@mbuf_body += sub_ex_d.get_decoded_body;
 			end
-##			@mbuf_body += "--#{@boundary}--\n";
-			@mbuf_body += "// ----------\n";
-
-		elsif @content_type_major == "text" && @encoding == "base64"
-			@mbuf_body_array.each do |buf|
-				tbuf += buf.unpack('m*')[0];
-			end
-		elsif @content_type_major == "text" && @encoding == "quoted-printable"
-			@mbuf_body_array.each do |buf|
-				tbuf += buf.unpack('M*')[0];
-			end
-		elsif @content_type_major == "text"
-			@mbuf_body_array.each do |buf|
-				tbuf += buf;
-			end
+			@mbuf_body += %Q|## <feml end-boundary="#{@boundary}" />\n|;
 
 		elsif @content_type_major == "message" &&
 		      @content_type_minor == "rfc822"
 			sub_ex_d = ExDecoder.new();
 			sub_ex_d.load_array(@mbuf_body_array);
-			@mbuf_body += "====================\n";
+			@mbuf_body += "## ====================\n";
 			@mbuf_body += sub_ex_d.get_decoded_body;
 			@mbuf_body += "// ====================\n";
 
-		elsif @content_type_major == "message"
-			@mbuf_body_array.each do |buf|
-				tbuf += buf;
+		elsif content_type_is_acceptable()
+			case @encoding
+			when "base64"
+				@mbuf_body_array.each do |buf|
+					tbuf += buf.unpack('m*')[0];
+				end
+			when "quoted-printable"
+				@mbuf_body_array.each do |buf|
+					tbuf += buf.unpack('M*')[0];
+				end
+			else
+				@mbuf_body_array.each do |buf|
+					tbuf += buf;
+				end
 			end
+			tbuf += "\n" if tbuf[-1, 1] != "\n";
 
 		else
 			### Ϥʤ
@@ -1044,8 +1128,7 @@ class Tokenizer
 	#         :add_spam_sub_clean | :add_clean_sub_spam
 	#
 	def update_occurences(mode = :add_clean)
-		dont_update = AmConf['DontUpdateTokenDB'];
-		return if dont_update != nil && ! dont_update.empty?;
+		return if is_conf_on('DontUpdateTokenDB');
 
 		db = PStore.new(AmConf['OccurenceStore']);
 		db.transaction do
@@ -1178,6 +1261,33 @@ def extract_mail_addr(buf)
 	return mail_addr;
 end
 
+# ================================================================
+# դν񼰤ȡϰäƤ뤫ɤǤ롣
+#
+def verify_date(date_line)
+	date_diag = '';				# դθ
+	date_std = '';				# ɸ񼰤ɽ
+	date_diff = 0;
+	begin
+		date_val = DateTime.parse(date_line);
+		date_now = DateTime::now;
+		date_diff = date_val - date_now;
+		date_std = date_val.new_offset(date_now.offset).strftime("%Y-%m-%d %H:%M:%S %z");
+	rescue ArgumentError
+		date_val = nil;
+		date_diag = "[] ORIG: #{date_line}";
+	else
+		if date_diff <= -1.5
+			date_diag = "[] (#{date_std})";
+		elsif 1.5 < date_diff
+			date_diag = "[̤] (#{date_std})";
+		else
+			date_diag = "";
+		end
+	end
+	return date_diag;
+end
+
 # ================================================================
 # ۥ磻ȡꥹ (ե˵) ˹פƤ뤫ݤθ
 # target_text: оʸ (̤ˤإåʸ)
@@ -1254,7 +1364,9 @@ def spam_filter(fp_in, user)
 	# stdinΥ᡼롦åɤ߹ߤʤ顢
 	# - Хååסե˽񤭽Ф
 	# - From סFrom:׹Ԥ᡼롦ɥ쥹ФƤ
-	# - Τ˶ǤʤԤФλݤϿ(ԤʤΤѴ)
+	# - Τ˶ǤʤԤ뤫ɤĴ٤롣
+	# - դɤĴ٤롣
+	# - ȤƤcharset󤷤Ƥ (multipartǤʣˤʤꤦ)
 	#
 	exd = ExDecoder.new();
 	exd.load_fp(fp_in);
@@ -1268,7 +1380,8 @@ def spam_filter(fp_in, user)
 
 	unix_from_line = exd.get_header('From ') || '';
 	from_header = exd.get_header('From:') || '';
-	subject = exd.get_header('Subject:') || "\n";
+	subject = exd.get_header('Subject:') || '';
+	date_header = exd.get_header('Date:') || '';
 	received = exd.get_header('Received:') || '';
 			# 'Received:' إåʣ硢1Ĥ᤬롣
 	mail_body = exd.get_decoded_body() || '';
@@ -1279,11 +1392,23 @@ def spam_filter(fp_in, user)
 	end
 	mail_address = extract_mail_addr(from_header);
 
+	date_diag = verify_date(date_header);
+
+	charset_used = [];
+	header_summary = exd.header_summary_line();
+	if %r|<feml.+charset="([^"]+)".+/>| =~ header_summary
+		charset_used << $1.downcase;
+	end
+
 	non_empty_line_found = false;
 	mail_body.each_line do |buf|
-		if /^\s*$/ !~ buf
+		if /^\s*$/ !~ buf && %r|<feml.+ />| !~ buf
 			non_empty_line_found = true;
 		end
+		if %r|<feml.+charset="([^"]+)".+/>| =~ buf
+			charset = $1.downcase;
+			charset_used << charset if ! charset_used.include?(charset);
+		end
 	end
 
 	# ================================================================
@@ -1292,24 +1417,37 @@ def spam_filter(fp_in, user)
 	sp = SpamProb.new(subject + "\n" + mail_body, lfp);
 	spam_prob = sp.probability;
 
+	spam_charsets_array = nil;
+	if (spam_charsets = AmConf['SpamCharsets']) != nil
+		spam_charsets_array = spam_charsets.split(/,\s*/);
+	end
+
 	result = '';
 	is_in_white_list = false;
+	spam_prob_percent = sprintf("%.2f%%", spam_prob * 100);
 	if is_in_white_list_pattern(unix_from, 'UfromWhiteList')
-		result = sprintf("(unix_from whitelist) (%f)", spam_prob);
+		result = "FemlWhite (unix_from) (#{spam_prob_percent})";
 		is_in_white_list = true;
 	elsif is_in_white_list_pattern(mail_address, 'FromWhiteList')
-		result = sprintf("(mail_from whitelist) (%f)", spam_prob);
+		result = "FemlWhite (mail_from) (#{spam_prob_percent})";
 		is_in_white_list = true;
 	elsif is_in_white_list_pattern(received, 'ReceivedWhiteList')
-		result = sprintf("(received: whitelist) (%f)", spam_prob);
+		result = "FemlWhite (received) (#{spam_prob_percent})";
 		is_in_white_list = true;
-	elsif ! non_empty_line_found
-		result = sprintf("(has only empty line) [%f]", spam_prob);
+	elsif is_conf_on('SpamIfEmptyBody') && ! non_empty_line_found
+		result = "FemlDirty (empty body) [#{spam_prob_percent}]";
+		spam_prob = 1.0;
+	elsif is_conf_on('SpamIfIllegalDate') && date_diag != ''
+		result = "FemlDirty (date line) [#{spam_prob_percent}] #{date_diag}";
+		spam_prob = 1.0;
+	elsif spam_charsets_array != nil &&
+	      ! (s_ch = (spam_charsets_array & charset_used).join(',')).empty?
+		result = "FemlDirty (charset: #{s_ch}) [#{spam_prob_percent}]";
 		spam_prob = 1.0;
 	elsif spam_prob < AmConf['SpamThreshold'].to_f
-		result = sprintf("S (%f)", spam_prob);
+		result = "FemlClean (#{spam_prob_percent})";
 	else
-		result = sprintf("S [%f]", spam_prob);
+		result = "FemlDirty [#{spam_prob_percent}]";
 	end
 
 	# ==============================================================
@@ -1326,8 +1464,11 @@ def spam_filter(fp_in, user)
 	#
 	lfp.print "** [LOCK] 㳰: #{rescue_point} \n" if lock_fp == nil;
 	lfp.print <<-"__LOG__";
-** #{timestamp} / #{result} / #{unix_from} / #{mail_address}
-** SubjectLine: #{subject}
+** #{timestamp} ----------------------------------------
+** #{result}
+** FromLine: #{mail_address} / UnixFrom: #{unix_from} / charset: #{charset_used.join(',')}
+** SubjectLine: #{subject.chomp}
+
 	__LOG__
 
 	lfp.close();
